import { getTargetHardwareUuid } from '../../core/app';
import { refreshDevice, AaiDevice } from '../../infrastructure';
import { CliTerseError } from '@alwaysai/alwayscli';
import { SshSpawner, logger, runWithSpinner } from '../../util';
import {
  DeviceCfgErrors,
  DeviceTokens,
  writeDeviceCfgFile,
  writeOrValidateDeviceCfgFile,
  writeTokens
} from '../../core/device';
import { REMOTE_AAI_CFG_DIR_LINUX } from '../../paths';
import { PLEASE_REPORT_THIS_ERROR_MESSAGE } from '../../constants';

export async function deviceCheckAndUpdateComponent(props: {
  yes: boolean;
  targetHostname: string;
  device: AaiDevice;
  isNewDevice: boolean;
}) {
  const { targetHostname, device, isNewDevice } = props;
  const spawner = SshSpawner({
    targetHostname,
    targetPath: REMOTE_AAI_CFG_DIR_LINUX
  });

  const CLEAR_OLD_CONF_MESSAGE =
    "To remove old configurations from the device run 'aai device clean', and then run 'aai app configure' again.";

  if (isNewDevice) {
    await runWithSpinner(
      async () => {
        try {
          await writeDeviceCfgFile({ spawner, deviceUuid: device.uuid });
        } catch (e) {
          if (e.code === DeviceCfgErrors.TARGET_DEVICE_CONFIGURED) {
            throw new CliTerseError(
              `The target device is already configured as an alwaysAI device! ${CLEAR_OLD_CONF_MESSAGE}`
            );
          } else {
            logger.error(e.message);
            throw new CliTerseError(
              `There was an error with configuring the device. ${PLEASE_REPORT_THIS_ERROR_MESSAGE}`
            );
          }
        }
      },
      [],
      'Configure device'
    );
  } else {
    await runWithSpinner(
      async () => {
        const hardwareId = await getTargetHardwareUuid(spawner);
        if (hardwareId !== device.hardware_ids) {
          throw new CliTerseError(
            `The target device does not match the selected device! ${CLEAR_OLD_CONF_MESSAGE}`
          );
        }
        try {
          await writeOrValidateDeviceCfgFile({
            spawner,
            deviceUuid: device.uuid
          });
        } catch (e) {
          if (e.code === DeviceCfgErrors.TARGET_DEVICE_MISMATCH) {
            throw new CliTerseError(
              `The target device does not match the selected device! ${CLEAR_OLD_CONF_MESSAGE}`
            );
          } else {
            logger.error(e.message);
            throw new CliTerseError(
              `There was an error with updating the device configuration. ${PLEASE_REPORT_THIS_ERROR_MESSAGE}`
            );
          }
        }
      },
      [],
      'Check device configuration'
    );
  }

  if (device.mode === 'development') {
    // Production tokens don't expire, and therefore don't need to be refreshed.
    // Additionally, we don't want to overwrite existing tokens on a provisioned
    // production device.
    await runWithSpinner(
      async () => {
        const { accessToken, refreshToken, idToken } = await refreshDevice(
          device.uuid
        );
        const tokens: DeviceTokens = {
          deviceId: device.uuid,
          accessToken,
          refreshToken,
          idToken
        };
        await writeTokens({ spawner, tokens });
      },
      [],
      'Write tokens to device'
    );
  }
}
